unit StringTree;

{$mode objfpc}{$H+}

// {.$DEFINE AVL_TREE}  -- defined in Project options

interface

uses
  Classes, SysUtils,
  {$IFDEF AVL_TREE}AVL_Tree{$ELSE}avglvltree{$ENDIF};

type
  { forward declarations }
  TsRowColAVLTree = class;

  { TsRowCol }
  TsRowCol = record
    Row, Col: Cardinal;
  end;
  PsRowCol = ^TsRowCol;

  { TsRowColEnumerator }
  TsRowColEnumerator = class
  private
  protected
    FCurrentNode: {$IFDEF AVL_TREE}TAVLTreeNode{$ELSE}TAvgLvlTreeNode{$ENDIF};
    FTree: TsRowColAVLTree;
    FStartRow, FEndRow, FStartCol, FEndCol: LongInt;
    FDone: Boolean;
    FReverse: Boolean;
    function GetCurrent: PsRowCol;
  public
    constructor Create(ATree: TsRowColAVLTree;
      AStartRow, AStartCol, AEndRow, AEndCol: Integer; AReverse: Boolean);
    function GetEnumerator: TsRowColEnumerator; inline;
    function MoveNext: Boolean;
    property Current: PsRowCol read GetCurrent;
    property StartRow: LongInt read FStartRow;
    property EndRow: LongInt read FEndRow;
    property StartCol: LongInt read FStartCol;
    property EndCol: LongInt read FEndCol;
  end;

  { TsRowColAVLTree }
  TsRowColAVLTree = class({$IFDEF AVL_TREE}TAVLTree{$ELSE}TAvgLvlTree{$ENDIF})
  private
    FOwnsData: Boolean;
  protected
    procedure DisposeData(var AData: Pointer); virtual; abstract;
    function NewData: Pointer; virtual; abstract;
  public
    constructor Create(AOwnsData: Boolean = true);
    destructor Destroy; override;
    function Add(ARow, ACol: Integer): PsRowCol; overload;
    procedure Clear;
    procedure Delete(ANode: {$IFDEF AVL_TREE}TAVLTreeNode{$ELSE}TAvgLvlTreeNode{$ENDIF}); overload;
    procedure Delete(ARow, ACol: Cardinal); overload;
    procedure DeleteRowOrCol(AIndex: Integer; IsRow: Boolean); virtual;
    procedure Exchange(ARow1, ACol1, ARow2, ACol2: Integer); virtual;
    function FindByRowCol(ARow, ACol: Integer): PsRowCol; overload;
    function GetData(ANode: {$IFDEF AVL_TREE}TAVLTreeNode{$ELSE}TAvgLvlTreeNode{$ENDIF}): PsRowCol;
    function GetFirst: PsRowCol;
    function GetLast: PsRowCol;
    procedure InsertRowOrCol(AIndex: Integer; IsRow: Boolean);
    procedure Remove(ARow, ACol: Integer); overload;
  end;

  { TsCells }
  TCell = record
    Row, Col: Integer;
    Value: String;
  end;
  PCell = ^TCell;

  TsCellEnumerator = class(TsRowColEnumerator)
  protected
    function GetCurrent: PCell;
  public
    function GetEnumerator: TsCellEnumerator; inline;
    property Current: PCell read GetCurrent;
  end;

  TsCells = class(TsRowColAVLTree)
  protected
    procedure DisposeData(var AData: Pointer); override;
    function NewData: Pointer; override;
  public
    function AddCell(ARow, ACol: Integer): PCell;
    procedure DeleteCell(ARow, ACol: Integer);
    function FindCell(ARow, ACol: Integer): PCell;
    function GetFirstCell: PCell;
    function GetFirstCellOfRow(ARow: Integer): PCell;
    function GetLastCell: PCell;
    function GetLastCellOfRow(ARow: Integer): PCell;
    // enumerators
    function GetEnumerator: TsCellEnumerator;
    function GetReverseEnumerator: TsCellEnumerator;
    function GetColEnumerator(ACol: Integer; AStartRow: Integer = 0;
      AEndRow: Integer = MaxInt): TsCellEnumerator;
    function GetRangeEnumerator(AStartRow, AStartCol,
      AEndRow, AEndCol: Integer): TsCellEnumerator;
    function GetRowEnumerator(ARow: Integer; AStartCol: Integer = 0;
      AEndCol: Integer = MaxInt): TsCellEnumerator;
    function GetReverseColEnumerator(ACol: Integer; AStartRow: Integer = 0;
      AEndRow: Integer = MaxInt): TsCellEnumerator;
    function GetReverseRangeEnumerator(AStartRow, AStartCol,
      AEndRow, AEndCol: Integer): TsCellEnumerator;
    function GetReverseRowEnumerator(ARow: Integer; AStartCol: Integer = 0;
      AEndCol: Integer = MaxInt): TsCellEnumerator;
  end;

  TsStringTable = class
  private
    FFirstRowIndex, FFirstColIndex: Integer;
    FLastRowIndex, FLastColIndex: Integer;
    function GetCells(ARow, ACol: Integer): String;
    function GetColCount: Integer;
    function GetRowCount: Integer;
    procedure SetCells(ARow, ACol: Integer; const AValue: String);
  protected
    function  AddCell(ARow, ACol: Integer): PCell;
  public
    FCells: TsCells;
    constructor Create;
    destructor Destroy; override;
    function  FindCell(ARow, ACol: Integer): PCell;
    function  GetCell(ARow, ACol: Integer): PCell;
    function  GetCellCount: Integer;
    function GetEnumerator: TsCellEnumerator;
    function GetFirstColIndex(AForceCalculation: Boolean = false): Integer;
    function GetFirstRowIndex(AForceCalculation: Boolean = false): Integer;
    function GetLastColIndex(AForceCalculation: Boolean = false): Integer;
    function GetLastRowIndex(AForceCalculation: Boolean = false): Integer;
    property Cells[ARow, ACol: Integer]: String read GetCells write SetCells;
    property ColCount: Integer read GetColCount;
    property RowCount: Integer read GetRowCount;
  end;


implementation

uses
  Math;


{ Helper function for sorting }

function CompareRowCol(Item1, Item2: Pointer): Integer;
begin
  Result := Longint(PsRowCol(Item1)^.Row) - PsRowCol(Item2)^.Row;
  if Result = 0 then
    Result := Longint(PsRowCol(Item1)^.Col) - PsRowCol(Item2)^.Col;
end;


{==============================================================================}
{                         TsRowColEnumerator                                   }
{ A specialized enumerator for TsRowColAVLTree using the  pointers to the data }
{ records.                                                                     }
{==============================================================================}

constructor TsRowColEnumerator.Create(ATree: TsRowColAVLTree;
  AStartRow, AStartCol, AEndRow, AEndCol: Integer; AReverse: Boolean);
var
  node: {$IFDEF AVL_TREE}TAVLTreeNode{$ELSE}TAvgLvlTreeNode{$ENDIF};
begin
  FTree := ATree;
  FReverse := AReverse;
  if AStartRow <= AEndRow then
  begin
    FStartRow := AStartRow;
    FEndRow := AEndRow;
  end else
  begin
    FStartRow := AEndRow;
    FEndRow := AStartRow;
  end;
  if AStartCol <= AEndCol then
  begin
    FStartCol := AStartCol;
    FEndCol := AEndCol;
  end else
  begin
    FStartCol := AEndCol;
    FEndCol := AStartCol;
  end;
  if FEndRow = $7FFFFFFF then
  begin
    node := FTree.FindHighest;
    if node <> nil then
      FEndRow := PsRowCol(node.Data)^.Row;
  end;
end;

function TsRowColEnumerator.GetCurrent: PsRowCol;
begin
  if Assigned(FCurrentNode) then
    Result := PsRowCol(FCurrentNode.Data)
  else
    Result := nil;
end;

function TsRowColEnumerator.GetEnumerator: TsRowColEnumerator;
begin
  Result := self;
end;

function TsRowColEnumerator.MoveNext: Boolean;
var
  curr: PsRowCol;
  rc: TsRowCol;
begin
  Result := false;
  if FCurrentNode <> nil then begin
    if FReverse then
    begin
      FCurrentNode := FTree.FindPrecessor(FCurrentNode);
      if FCurrentNode <> nil then
      begin
        curr := PsRowCol(FCurrentNode.Data);
        if not InRange(LongInt(curr^.Col), FStartCol, FEndCol) then
        begin
          rc := curr^;
          if LongInt(rc.Col) < FStartCol then
            dec(LongInt(rc.Row));
          rc.Col := FEndCol;
          FCurrentNode := FTree.FindNearest(@rc);
          if FCurrentNode <> nil then begin
            curr := PsRowCol(FCurrentNode.Data);
            while (FCurrentNode <> nil) and
                  not (InRange(curr^.Row, FStartRow, FEndRow) and InRange(curr^.Col, FStartCol, FEndCol))
            do begin
              FCurrentNode := FTree.FindPrecessor(FCurrentNode);
              if FCurrentNode <> nil then curr := PsRowCol(FCurrentNode.Data);
            end;
          end;
        end;
      end;
    end else
    begin
      FCurrentNode := FTree.FindSuccessor(FCurrentNode);
      if FCurrentNode <> nil then
      begin
        curr := PsRowCol(FCurrentNode.Data);
        rc.Col := FStartCol;
        if LongInt(rc.Col) > FEndCol then inc(rc.Row);
        if not InRange(LongInt(curr^.Col), FStartCol, FEndCol) then
        begin
          rc := curr^;
          if LongInt(rc.Col) > FEndCol then inc(rc.Row);
          rc.Col := FStartCol;
          FCurrentNode := FTree.FindNearest(@rc);
          if FCurrentNode <> nil then
          begin
            curr := PsRowCol(FCurrentNode.Data);
            if (LongInt(curr^.Col) < FStartCol) then
              while (FCurrentNode <> nil) and not InRange(curr^.Col, FStartCol, FEndCol) do
              begin
                FCurrentNode := FTree.FindSuccessor(FCurrentNode);
                if FCurrentNode <> nil then curr := PsRowCol(FCurrentNode.Data);
              end;
            while (FCurrentNode <> nil) and
                  not (InRange(curr^.Row, FStartRow, FEndRow) and InRange(curr^.Col, FStartCol, FEndCol))
            do begin
              FCurrentNode := FTree.FindSuccessor(FCurrentNode);
              if FCurrentNode <> nil then curr := PsRowCol(FCurrentNode.Data);
            end;
          end;
        end;
      end;
    end;
    Result := (FCurrentNode <> nil) and
              InRange(curr^.Col, FStartCol, FEndCol) and
              InRange(curr^.Row, FStartRow, FEndRow);
  end else
  begin
    if FReverse then
    begin
      rc.Row := FEndRow;
      rc.Col := FEndCol;
      FCurrentNode := FTree.FindNearest(@rc);
      if FCurrentNode <> nil then
        curr := PsRowCol(FCurrentNode.Data);
      while (FCurrentNode <> nil) and
            not (InRange(curr^.Row, FStartRow, FEndRow) and InRange(curr^.Col, FStartCol, FEndCol))
      do begin
        FCurrentNode := FTree.FindPrecessor(FCurrentNode);
        if FCurrentNode <> nil then curr := PsRowCol(FCurrentNode.Data);
      end;
    end else
    begin
      rc.Row := FStartRow;
      rc.Col := FStartCol;
      FCurrentNode := FTree.FindNearest(@rc);
      if FCurrentNode <> nil then
        curr := PsRowCol(FCurrentNode.Data);
      while (FCurrentNode <> nil) and
            not (InRange(curr^.Row, FStartRow, FEndRow) and InRange(curr^.Col, FStartCol, FEndCol))
      do begin
        FCurrentNode := FTree.FindSuccessor(FCurrentNode);
        if FCurrentNode <> nil then curr := PsRowCol(FCurrentNode.Data);
      end;
    end;
    Result := (FCurrentNode <> nil);
  end;
end;



{@@ ----------------------------------------------------------------------------
  Constructor of the AVLTree. Installs a compare procedure for row and column
  indexes. If AOwnsData is true then the tree automatically destroys the
  data records attached to the tree nodes.
-------------------------------------------------------------------------------}
constructor TsRowColAVLTree.Create(AOwnsData: Boolean = true);
begin
  inherited Create(@CompareRowCol);
  FOwnsData := AOwnsData;
end;

{@@ ----------------------------------------------------------------------------
  Destructor of the AVLTree. Clears the tree nodes and, if the tree has been
  created with AOwnsData=true, destroys the data records
-------------------------------------------------------------------------------}
destructor TsRowColAVLTree.Destroy;
begin
  Clear;
  inherited;
end;

{@@ ----------------------------------------------------------------------------
  Adds a new node to the tree identified by the specified row and column
  indexes.
-------------------------------------------------------------------------------}
function TsRowColAVLTree.Add(ARow, ACol: Integer): PsRowCol;
begin
  Result := NewData;
  Result^.Row := ARow;
  Result^.Col := ACol;
  inherited Add(Result);
end;

{@@ ----------------------------------------------------------------------------
  Clears the tree, i.e, destroys the data records (if the tree has been created
  with AOwnsData = true) and removes all nodes.
-------------------------------------------------------------------------------}
procedure TsRowColAVLTree.Clear;
var
  node, nextnode: {$IFDEF AVL_TREE}TAVLTreeNode{$ELSE}TAvgLvlTreeNode{$ENDIF};
begin
  node := FindLowest;
  while node <> nil do begin
    nextnode := FindSuccessor(node);
    Delete(node);
    node := nextnode;
  end;
end;

{@@ ----------------------------------------------------------------------------
  Removes the specified node from the tree. If the tree has been created with
  AOwnsData = true then the data record is destroyed as well
-------------------------------------------------------------------------------}
procedure TsRowColAVLTree.Delete(ANode: {$IFDEF AVL_TREE}TAVLTreeNode{$ELSE}TAvgLvlTreeNode{$ENDIF});
begin
  if FOwnsData and Assigned(ANode) then
    DisposeData(PsRowCol(ANode.Data));
  inherited Delete(ANode);
end;

procedure TsRowColAVLTree.Delete(ARow, ACol: Cardinal);
var
  node: {$IFDEF AVL_TREE}TAVLTreeNode{$ELSE}TAvgLvlTreeNode{$ENDIF};
  cell: TCell;
begin
  cell.Row := ARow;
  cell.Col := ACol;
  node := inherited Find(@cell);
  if Assigned(node) then
    Delete(node);
end;

{@@ ----------------------------------------------------------------------------
  This procedure adjusts row or column indexes stored in the tree nodes if a
  row or column will be deleted from the underlying worksheet.

  @param  AIndex  Index of the row (if IsRow=true) or column (if IsRow=false)
                  to be deleted
  @param  IsRow   Identifies whether AIndex refers to a row or column index
-------------------------------------------------------------------------------}
procedure TsRowColAVLTree.DeleteRowOrCol(AIndex: Integer; IsRow: Boolean);
var
  node, nextnode: {$IFDEF AVL_TREE}TAVLTreeNode{$ELSE}TAvgLvlTreeNode{$ENDIF};
  item: PsRowCol;
begin
  node := FindLowest;
  while Assigned(node) do begin
    nextnode := FindSuccessor(node);
    item := PsRowCol(node.Data);
    if IsRow then
    begin
      // Update all RowCol records at row indexes above the deleted row
      if item^.Row > AIndex then
        dec(item^.Row)
      else
      // Remove and destroy the RowCol record if it is in the deleted row
      if item^.Row = AIndex then
        Delete(node);
    end else
    begin
      // Update all RowCol records at column indexes above the deleted column
      if item^.Col > AIndex then
        dec(item^.Col)
      else
      // Remove the RowCol record if it is in the deleted column
      if item^.Col = AIndex then
        Delete(node);
    end;
    node := nextnode;
  end;
end;

{@@ ----------------------------------------------------------------------------
  Exchanges two nodes
-------------------------------------------------------------------------------}
procedure TsRowColAVLTree.Exchange(ARow1, ACol1, ARow2, ACol2: Integer);
var
  item1, item2: PsRowCol;
begin
  item1 := FindByRowCol(ARow1, ACol1);
  item2 := FindByRowCol(ARow2, ACol2);

  // There are entries for both locations: Exchange row/col indexes
  if (item1 <> nil) and (item2 <> nil) then
  begin
    Remove(item1);
    Remove(item2);
    item1^.Row := ARow2;
    item1^.Col := ACol2;
    item2^.Row := ARow1;
    item2^.Col := ACol1;
    inherited Add(item1);   // The items are sorted to the correct position
    inherited Add(item2);   // when they are added to the tree
  end else

  // Only the 1st item exists --> give it the row/col indexes of the 2nd item
  if (item1 <> nil) then
  begin
    Remove(item1);
    item1^.Row := ARow2;
    item1^.Col := ACol2;
    inherited Add(item1);
  end else

  // Only the 2nd item exists --> give it the row/col indexes of the 1st item
  if (item2 <> nil) then
  begin
    Remove(item2);
    item2^.Row := ARow1;
    item2^.Col := ACol1;
    inherited Add(item2);  // just adds the existing item at the new position
  end;
end;

{@@ ----------------------------------------------------------------------------
  Seeks the entire tree for a node of the specified row and column indexes and
  returns a pointer to the data record.
  Returns nil if such a node does not exist
-------------------------------------------------------------------------------}
function TsRowColAVLTree.FindByRowCol(ARow, ACol: Integer): PsRowCol;
var
  data: TsRowCol;
  node: {$IFDEF AVL_TREE}TAVLTreeNode{$ELSE}TAvgLvlTreeNode{$ENDIF};
begin
  Result := nil;
  if  (Count = 0) then
    exit;

  data.Row := ARow;
  data.Col := ACol;
  node := inherited Find(@data);
  if Assigned(node) then
    Result := PsRowCol(node.Data);
end;

{@@ ----------------------------------------------------------------------------
  Extracts the pointer to the data record from a tree node
-------------------------------------------------------------------------------}
function TsRowColAVLTree.GetData(
  ANode: {$IFDEF AVL_TREE}TAVLTreeNode{$ELSE}TAvgLvlTreeNode{$ENDIF}): PsRowCol;
begin
  if ANode <> nil then
    Result := PsRowCol(ANode.Data)
  else
    Result := nil;
end;

{@@ ----------------------------------------------------------------------------
  The combination of the methods GetFirst and GetNext allow a fast iteration
  through all nodes of the tree.
-------------------------------------------------------------------------------}
function TsRowColAVLTree.GetFirst: PsRowCol;
begin
  Result := GetData(FindLowest);
end;

function TsRowColAVLTree.GetLast: PsRowCol;
begin
  Result := GetData(FindHighest);
end;

{@@ ----------------------------------------------------------------------------
  This procedure adjusts row or column indexes stored in the tree nodes if a
  row or column will be inserted into the underlying worksheet.

  @param  AIndex  Index of the row (if IsRow=true) or column (if IsRow=false)
                  to be inserted
  @param  IsRow   Identifies whether AIndex refers to a row or column index
-------------------------------------------------------------------------------}
procedure TsRowColAVLTree.InsertRowOrCol(AIndex: Integer; IsRow: Boolean);
var
  node: {$IFDEF AVL_TREE}TAVLTreeNode{$ELSE}TAvgLvlTreeNode{$ENDIF};
  item: PsRowCol;
begin
  node := FindLowest;
  while Assigned(node) do begin
    item := PsRowCol(node.Data);
    if IsRow then
    begin
      if item^.Row >= AIndex then inc(item^.Row);
    end else
    begin
      if item^.Col >= AIndex then inc(item^.Col);
    end;
    node := FindSuccessor(node);
  end;
end;

{@@ ----------------------------------------------------------------------------
  Removes the node, but does NOT destroy the associated data reocrd
-------------------------------------------------------------------------------}
procedure TsRowColAVLTree.Remove(ARow, ACol: Integer);
var
  node: {$IFDEF AVL_TREE}TAVLTreeNode{$ELSE}TAvgLvlTreeNode{$ENDIF};
  item: TsRowCol;
begin
  item.Row := ARow;
  item.Col := ACol;
  node := inherited Find(@item);
  Remove(node);
//  Delete(node);
end;


{==============================================================================}
{ TsCellEnumerator: enumerator for the TsCells AVLTree                        }
{==============================================================================}

function TsCellEnumerator.GetEnumerator: TsCellEnumerator;
begin
  Result := self;
end;

function TsCellEnumerator.GetCurrent: PCell;
begin
  Result := PCell(inherited GetCurrent);
end;


{==============================================================================}
{ TsCells: an AVLTree to store spreadsheet cells                               }
{==============================================================================}

{@@ ----------------------------------------------------------------------------
  Adds a node with a new TCell record to the tree.
  Returns a pointer to the cell record.
  NOTE: It must be checked first that there ia no other record at the same
        col/row. (Check omitted for better performance).
-------------------------------------------------------------------------------}
function TsCells.AddCell(ARow, ACol: Integer): PCell;
begin
  Result := PCell(Add(ARow, ACol));
end;

{@@ ----------------------------------------------------------------------------
  Deletes the node for the specified row and column index along with the
  associated cell data record.
-------------------------------------------------------------------------------}
procedure TsCells.DeleteCell(ARow, ACol: Integer);
begin
  Delete(ARow, ACol);
end;

{@@ ----------------------------------------------------------------------------
  Helper procedure which disposes the memory occupied by the cell data
  record attached to a tree node.
-------------------------------------------------------------------------------}
procedure TsCells.DisposeData(var AData: Pointer);
begin
  if AData <> nil then
    Dispose(PCell(AData));
  AData := nil;
end;

{@@ ----------------------------------------------------------------------------
  Checks whether a specific cell already exists
-------------------------------------------------------------------------------}
function TsCells.FindCell(ARow, ACol: Integer): PCell;
begin
  Result := PCell(FindByRowCol(ARow, ACol));
end;

{@@ ----------------------------------------------------------------------------
  Cell enumerators (use in "for ... in" syntax)
-------------------------------------------------------------------------------}
function TsCells.GetEnumerator: TsCellEnumerator;
begin
  Result := TsCellEnumerator.Create(self, 0, 0, MaxInt, MaxInt, false);
end;

function TsCells.GetColEnumerator(ACol: Integer; AStartRow: Integer = 0;
  AEndRow: Integer = MaxInt): TsCellEnumerator;
begin
  Result := TsCellEnumerator.Create(Self, AStartRow, ACol, AEndRow, ACol, false);
end;

function TsCells.GetRangeEnumerator(AStartRow, AStartCol,
  AEndRow, AEndCol: Integer): TsCellEnumerator;
begin
  Result := TsCellEnumerator.Create(Self,
    AStartRow, AStartCol, AEndRow, AEndCol, false);
end;

function TsCells.GetRowEnumerator(ARow: Integer; AStartCol: Integer = 0;
  AEndCol: Integer = MaxInt): TsCellEnumerator;
begin
  Result := TsCellEnumerator.Create(Self,
    ARow, AStartCol, ARow, AEndCol, false);
end;

function TsCells.GetReverseColEnumerator(ACol: Integer; AStartRow: Integer = 0;
  AEndRow: Integer = $7FFFFFFF): TsCellEnumerator;
begin
  Result := TsCellEnumerator.Create(Self,
    AStartRow, ACol, AEndRow, ACol, true);
end;

function TsCells.GetReverseEnumerator: TsCellEnumerator;
begin
  Result := TsCellEnumerator.Create(self, 0, 0, $7FFFFFFF, $7FFFFFFF, true);
end;

function TsCells.GetReverseRangeEnumerator(AStartRow, AStartCol,
  AEndRow, AEndCol: Integer): TsCellEnumerator;
begin
  Result := TsCellEnumerator.Create(Self,
    AStartRow, AStartCol, AEndRow, AEndCol, true);
end;

function TsCells.GetReverseRowEnumerator(ARow: Integer; AStartCol: Integer = 0;
  AEndCol: Integer = $7FFFFFFF): TsCellEnumerator;
begin
  Result := TsCellEnumerator.Create(Self,
    ARow, AStartCol, ARow, AEndCol, true);
end;

{@@ ----------------------------------------------------------------------------
  Returns a pointer to the first cell of the tree.
-------------------------------------------------------------------------------}
function TsCells.GetFirstCell: PCell;
begin
  Result := PCell(GetFirst);
end;

{@@ ----------------------------------------------------------------------------
  Returns a pointer to the first cell in a specified row
-------------------------------------------------------------------------------}
function TsCells.GetFirstCellOfRow(ARow: Integer): PCell;
begin
  Result := nil;
  // Creating the row enumerator automatically finds the first cell of the row
  for Result in GetRowEnumerator(ARow) do
    exit;
end;

{@@ ----------------------------------------------------------------------------
  Returns a pointer to the last cell of the tree.
-------------------------------------------------------------------------------}
function TsCells.GetLastCell: PCell;
begin
  Result := PCell(GetLast);
end;

{@@ ----------------------------------------------------------------------------
  Returns a pointer to the last cell of a specified row
-------------------------------------------------------------------------------}
function TsCells.GetLastCellOfRow(ARow: Integer): PCell;
begin
  Result := nil;
  // Creating the reverse row enumerator finds the last cell of the row
  for Result in GetReverseRowEnumerator(ARow) do
    exit;
end;

{@@ ----------------------------------------------------------------------------
  Alloates memory for a cell data record.
-------------------------------------------------------------------------------}
function TsCells.NewData: Pointer;
var
  cell: PCell;
begin
  New(cell);
  cell^.Row := -1;
  cell^.Col := -1;
  cell^.Value := '';
  Result := cell;
end;


{==============================================================================}
{ TsStringTable: a 2D sparse matrix of strings                                 }
{==============================================================================}

constructor TsStringTable.Create;
begin
  inherited;
  FCells := TsCells.Create(true);
  FFirstRowIndex := -1;
  FLastRowIndex := -1;
  FFirstColIndex := -1;
  FLastColIndex := -1;
end;

destructor TsStringTable.Destroy;
begin
  FCells.Free;
  inherited;
end;

{ Creates a new cell
  NOTE: It is not checked if there exists already another cell at this location.
  This case must be avoided. USE CAREFULLY WITHOUT FindCell }
function TsStringTable.AddCell(ARow, ACol: Integer): PCell;
begin
  Result := FCells.AddCell(ARow, ACol);

  if FFirstColIndex = -1 then
    FFirstColIndex := GetFirstColIndex(true) else
    FFirstColIndex := Min(FFirstColIndex, ACol);

  if FFirstRowIndex = -1 then
    FFirstRowIndex := GetFirstRowIndex(true) else
    FFirstRowIndex := Min(FFirstRowIndex, ARow);

  if FLastColIndex = -1 then
    FLastColIndex := GetLastColIndex(true) else
    FLastColIndex := Max(FLastColIndex, ACol);

  if FLastRowIndex = -1 then
    FLastRowIndex := GetLastRowIndex(true) else
    FLastRowIndex := Max(FLastRowIndex, ARow);
end;

{ Finds the cell at (ARow, ACol). Returns nil if the cell does not exist. }
function TsStringTable.FindCell(ARow, ACol: Integer): PCell;
begin
  Result := PCell(FCells.FindByRowCol(ARow, ACol));
end;

{ Finds the cell at (ARow, ACol). If not found an empty cell is created.
  --> After calling GetCell it can safely be assumed that the cell exists. }
function TsStringTable.GetCell(ARow, ACol: Integer): PCell;
begin
  Result := FCells.FindCell(ARow, ACol);
  if Result = nil then
    Result := AddCell(ARow, ACol);
end;

function TsStringTable.GetCellCount: Integer;
begin
  Result := FCells.Count;
end;

{ Getter for the property Cells }
function TsStringTable.GetCells(ARow, ACol: Integer): String;
var
  cell: PCell;
begin
  cell := FindCell(ARow, ACol);
  if cell <> nil then Result := cell^.Value else Result := '';
end;

function TsStringTable.GetColCount: Integer;
begin
  Result := GetLastColIndex + 1;
end;

function TsStringTable.GetEnumerator: TsCellEnumerator;
begin
  Result := TsCellEnumerator.Create(FCells, 0, 0, MaxInt, MaxInt, false);
end;

function TsStringTable.GetFirstColIndex(AForceCalculation: Boolean = false): Integer;
var
  cell: PCell;
  i: Integer;
begin
  if AForceCalculation then
  begin
    if FCells.Count > 0 then
    begin
      Result := MaxInt;
      for cell in FCells do
        Result := Math.Min(Result, cell^.Col);
    end else
      Result := -1;
    // Store the result
    FFirstColIndex := Result;
  end
  else
  begin
    Result := FFirstColIndex;
    if Result = -1 then
      Result := GetFirstColIndex(true);
  end;
end;

function TsStringTable.GetFirstRowIndex(AForceCalculation: Boolean = false): Integer;
var
  cell: PCell;
  i: Integer;
begin
  if AForceCalculation then
  begin
    Result := -1;
    cell := FCells.GetFirstCell;
    if cell <> nil then Result := cell^.Row;
    FFirstRowIndex := Result;
  end
  else
  begin
    Result := FFirstRowIndex;
    if Result = -1 then
      Result := GetFirstRowIndex(true);
  end;
end;

function TsStringTable.GetLastColIndex(AForceCalculation: Boolean = false): Integer;
var
  cell: PCell;
begin
  if AForceCalculation or (FLastColIndex = -1) then
  begin
    // Traverse the tree from lowest to highest.
    // Since tree primary sort order is on row highest col could exist anywhere.
    Result := -1;
    for cell in FCells do
      Result := Math.Max(Result, cell^.Col);
    // Store the result
    FLastColIndex := Result;
  end
  else
    Result := FLastColIndex;
end;

function TsStringTable.GetLastRowIndex(AForceCalculation: Boolean = false): Integer;
var
  cell: PCell;
begin
  if AForceCalculation or (FLastRowIndex = -1) then
  begin
    // Index of highest row with at least one existing cell
    Result := -1;
    cell := FCells.GetLastCell;
    if Assigned(cell) then
      Result := cell^.Row;
    FLastRowIndex := Result;
  end
  else
    Result := FLastRowIndex
end;

function TsStringTable.GetRowCount: Integer;
begin
  Result := GetLastRowIndex + 1;
end;

{ Setter for the property Cells }
procedure TsStringTable.SetCells(ARow, ACol: Integer; const AValue: String);
var
  cell: PCell;
begin
  cell := GetCell(ARow, ACol);
  cell^.Value := AValue;
end;


end.

